﻿EnableExplicit

; Simple DNS Proxy
;	2017			(c) Luna Sole

;{ OnError }
	
	CompilerIf Not #PB_Compiler_Debugger
		Procedure ErrorHandler()
			Protected DieMsg$ = GetFilePart(ErrorFile()) + " [" + ErrorLine() + "]" + #CRLF$ + 
			                    UCase(ErrorMessage()) + #CRLF$ + RSet("", 64, " ")
			MessageRequester("Critical shit", DieMsg$, #PB_MessageRequester_Error | #PB_MessageRequester_Ok)
		EndProcedure
		OnErrorCall(@ErrorHandler())
	CompilerEndIf
	
;}


;{ Declares }

	Declare LogPrint(s$)

;}

;{ Common }
	
	; Inverts variable byte order (BigEndian <> LittleEndian)
	; RETURN:		inverted val
	Procedure SwapEndian(val)
		!MOV eax, dword[p.v_val]
		!BSWAP eax
		ProcedureReturn
	EndProcedure
	Procedure.w SwapEndianW(valW.w)
		Protected l.l = SwapEndian(valW)
		ProcedureReturn PeekW(@L + 2)
	EndProcedure
	
;}


;{ AppData }
	
	; https://technet.microsoft.com/en-us/library/dd197470(v=ws.10).aspx
	Structure DNS_HEADER
		TransactionID.w		; A 16-bit field identifying a specific DNS transaction. The transaction ID is created by the message originator and is copied by the responder into its response message. Using the transaction ID, the DNS client can match responses to its requests.
		Flags.w				; A 16-bit field containing various service flags that are communicated between the DNS client and the DNS server
		NumEntriesQ.w		; A 16-bit field representing the number of entries in the question section of the DNS message.
		NumEntriesA.w		; A 16-bit field representing the number of entries in the answer section of the DNS message.
		ARRc.w				; A 16-bit field representing the number of authority resource records in the DNS message.
		RRCex.w				; A 16-bit field representing the number of additional resource records in the DNS message.
	EndStructure
	
	; max size of UDP packet (and max supported here size of DNS query also)
	#UDP_MAX = 2048
	
	; type used to store info about clients
	Structure CLIENT_INFO
		connection.i			; UDP connection ID
		real_TransactionID.w	; DNS header transaction ID received from client
		date.l					; Timestamp of recent client activity
	EndStructure
	; used to store connections info
	Structure CONNECT_INFO
		ID.i					; UDP connection ID
		ref.w					; number of client queries using this connection
	EndStructure
	
	; global storages
	NewMap connections.CONNECT_INFO()		; connection id/EventClient() used as key
	NewMap clients.CLIENT_INFO()			; transaction ID (to be exact, fake ID produced by proxy) used as key
	; used to cleanup clients by timeout
	Global Cycle = Date()
	
	; Misc app data
	Global InitOK
	#APP_TITLE = "DNS Proxy 1.0.0." + #PB_EDITOR_Buildcount

	; DLL filter function prototype
	Prototype FilterQuery(Domain$)
	Global DLL_FilterQuery.FilterQuery
	
;}


;{ DNS proxy }
	
	; Process received DNS message from client (client is any program using DNS resolve on machine)
	; Out()		map to store global client data
	; OutC()	map with global connections pool
	; *Data		memory containing raw DNS message
	; len		actual message size
	; hClien	client connection ID
	; hTrueDNS	a DNS server connection to forward message
	; RETURN:	none
	Procedure DNSProcessQ(Map Out.CLIENT_INFO(), Map OutC.CONNECT_INFO(), *Data.DNS_HEADER, len, hClient, hTrueDNS)
		If len < SizeOf(DNS_HEADER)
			; there is nothing to do
			ProcedureReturn 
		EndIf
		
		; a list of domains queried to resolve
		Protected NewList Domains$()	
		
		; now read domain name[s]
		Protected Rq = Abs(SwapEndianW(*data\NumEntriesQ))
		; walk through all query records
		Define *p.Ascii = *data + SizeOf(DNS_HEADER)	
		While Rq
			; parse domain strings
			Define NStr = 0
			Define t$ = ""
			Repeat
				If (*p >= (*data + len))
					; wtf bitches going to overrun that buffer
					t$ = ""
					Rq = 0
					Break
				EndIf
				
				If *p\a = 0
					; should be end of string (record)
					Rq - 1
					; if all fine, add domain to list
					If t$
						AddElement(Domains$())
						Domains$() = t$
					EndIf
					*p + 1
					Break
				Else
					If NStr = 0
						; read len
						NStr = *p\a 
						If t$
							t$ + "."
						EndIf
					Else
						; reading string
						NStr - 1
						t$ + Chr(*p\a)
					EndIf
					*p + 1
				EndIf
			ForEver
		Wend
		
		
		; here can do something with domains (block resolving, or spoof IP if write some more code)
		Define tD$ = ""
		Define stopD$ = ""
		Define Action = 0
		ForEach Domains$()
			; ask filter DLL for action, if it is loaded
			If Action = 0 And DLL_FilterQuery
				Action = DLL_FilterQuery(Domains$())
				stopD$ = Domains$()
			EndIf
			
			; add domain to a string list
			If tD$
				tD$ + ", " 
			EndIf
			tD$ + Domains$()
		Next
		; filter DLL should return something between 0 and 2
		If Action < 0 Or Action > 2
			Action = 0
		EndIf
		
		; do what decided
		Select Action
			Case 0 ; 0 or anything else: all is OK, send resolve request to a true DNS server
				; this is fake TransactionID used internally
				; it must be unique within all current clients
				Static FakeID.w 
				FakeID + 1
				; true ID will be stored in client info map and used when sending responce back to client
				Protected TrueID.w = *data\TransactionID
				; patch transaction ID
				*data\TransactionID = FakeID
		
				; redirect patched query to a true DNS server
				If SendNetworkData(hTrueDNS, *data, len) = len
					; If all OK, store client info for future
					Out(Str(FakeID))\connection = hClient
					Out(Str(FakeID))\real_TransactionID = TrueID
					Out(Str(FakeID))\date = Date()
					; Add connection ref
					OutC(Str(hClient))\ref + 1
					OutC(Str(hClient))\ID = hClient
					; success notify
					LogPrint ("Q  [" + RSet(Hex(SwapEndianW(TrueID), #PB_Word), 4, "0") + "] : " + tD$)
				EndIf
				
			Case 1: ; block resolving
				; close current connection if it has 0 refs
				If OutC(Str(hClient))\ref = 0
		 			CloseNetworkConnection(hClient)
		 			DeleteMapElement(OutC())
		 		EndIf
				LogPrint ("Q  [" + RSet(Hex(SwapEndianW(*data\TransactionID), #PB_Word), 4, "0") + "] : block (" + stopD$ + ")")
							
			Case 2: ; block resolving + send empty DNS header to a client. that's OK to not make client wait until timeout expires (which happens if only blocked query)
				; send shit back
				*Data\NumEntriesA = 0
				*Data\NumEntriesQ = 0
				*Data\RRCex = 0
				*Data\Flags = 0
				SendNetworkData(hClient, *data, SizeOf(DNS_HEADER))
				
				; close current connection if it has 0 refs
				If OutC(Str(hClient))\ref = 0
		 			CloseNetworkConnection(hClient)
		 			DeleteMapElement(OutC())
		 		EndIf
				LogPrint ("Q  [" + RSet(Hex(SwapEndianW(*data\TransactionID), #PB_Word), 4, "0") + "] : block-F (" + stopD$ + ")")
				
		EndSelect
		
		; cls
		ClearList(Domains$())
		
	EndProcedure
	
	
	; Process data from DNS server and send it to corresponding client "as is"
	; Out()		map to get global client data
	; OutC()	map with global connections pool
	; *Data		memory containing raw DNS message
	; len		actual message size
	; RETURN:	none
	Procedure DNSProcessA(Map Out.CLIENT_INFO(), Map OutC.CONNECT_INFO(), *Data.DNS_HEADER, len)
		If len < SizeOf(DNS_HEADER)
			; there is nothing to do with header only or even less
			ProcedureReturn 
		EndIf
		

		Protected.w FakeClientID, RealClientID
		; current ID is fake (used by proxy to identify clients)
		FakeClientID = *Data\TransactionID
		; search for client in storage
		If FindMapElement(Out(), Str(FakeClientID))
			; patch ID back to a true ID
			RealClientID = Out()\real_TransactionID
			*Data\TransactionID = RealClientID

			; return responce from true server to a client
			If SendNetworkData(Out()\connection, *Data, len) = len
				; success notify
				LogPrint ("A  [" + RSet(Hex(SwapEndianW(RealClientID), #PB_Word), 4, "0") + "] = " + Str(len) + "b")
			EndIf
			; cls
			; dec connection ref
			OutC(Str(out()\connection))\ref - 1
			If OutC()\ref = 0
	 			CloseNetworkConnection(OutC()\ID)
	 			DeleteMapElement(OutC())
	 		EndIf
			DeleteMapElement(Out())
		EndIf
	EndProcedure
	
	
;}


;{ UI }

	;{ Dialogs }
	
		Global hF = LoadFont(#PB_Any, "Consolas", 10)
		SetGadgetFont(-1, FontID(hF))
		; UI callbacks
		; The following variables usable with DialogGadget() and related functions
		Global dlg_wndMain = CreateDialog(#PB_Any)
	
		; Generated with DialogDesigner2 [1.0.0.7], 2017/07/31 02:19:53
		Global DialogsXML$ = #Empty$
		DialogsXML$ +
			"<dialogs>" +
			"	<window name='wndMain' flags='#PB_Window_SystemMenu|#PB_Window_SizeGadget|#PB_Window_ScreenCentered' minwidth='auto' minheight='auto'>" +
			"		<hbox name='hbox1' expand='item:1'>" +
			"			<editor name='txtLog' flags='#PB_Editor_ReadOnly|#PB_Editor_WordWrap' height='320' width='512'/>" +
			"			<vbox name='vbox1' width='60' align='center' expand='item:3'>" +
			"				<button name='button1' flags='#PB_Button_Left' text='1'/>" +
			"				<button name='button2' flags='#PB_Button_Right' text='2'/>" +
			"				<container name='container1' disabled='yes'>" +
			"					<editor name='editor2' disabled='no' flags='#PB_Editor_WordWrap' text='DNS Pro xy by Lu na so le' invisible='no'/>" +
			"				</container>" +
			"			</vbox>" +
			"		</hbox>" +
			"	</window>" +
			"</dialogs>"
	
	;}


	; write UI log and file log
	Procedure LogPrint (s$)
		s$ = FormatDate("%hh:%ii:%ss | ", Date()) + s$
		Static logFile
		If IsFile(logFile) = 0
			logFile = CreateFile(#PB_Any, "log.log", #PB_File_SharedRead | #PB_UTF8)
			If IsFile(logFile)
				WriteStringFormat(logFile, #PB_UTF8)
			EndIf
		EndIf
		
		Protected l = DialogGadget(dlg_wndMain, "txtLog")
		If CountGadgetItems(l) > 1280
			RemoveGadgetItem(l, 0)
		EndIf
		AddGadgetItem(l, -1, s$)
		
		; scrol to the bottom
		Protected LenT.GETTEXTLENGTHEX
		Protected S.CHARRANGE
		S\cpMin = SendMessage_(GadgetID(l), #EM_GETTEXTLENGTHEX, LenT, 0)
		S\cpMax = S\cpMin
		SendMessage_(GadgetID(l), #EM_EXSETSEL, 0, S)
		
		If IsFile(logFile)
			WriteStringN(logFile, s$)
		EndIf
	EndProcedure
	
	
	; Parse dialogs XML
	Define WndMain
	Define DialogsXML = ParseXML(#PB_Any, DialogsXML$)
	; Open dialogs, bind events, customize
	If OpenXMLDialog(dlg_wndMain, DialogsXML, "wndMain")
		WndMain = DialogWindow(dlg_wndMain)
		SetWindowTitle(WndMain, #APP_TITLE)
		
		GadgetToolTip(DialogGadget(dlg_wndMain, "button1"), ~"does nothing")
		GadgetToolTip(DialogGadget(dlg_wndMain, "button2"), ~"does nothing")
		RefreshDialog(dlg_wndMain)
	EndIf
	; Cls
	If IsXML(DialogsXML): FreeXML(DialogsXML): EndIf
	DialogsXML$ = #Empty$

;}

;{ Init }
	
	InitOK = #True
	
	; start
	If InitNetwork() = 0
		LogPrint ("Failed to init network runtime")
		InitOK = #False
	EndIf
	
	; create server to listen for clients
	Define hServ = CreateNetworkServer(#PB_Any, 53, #PB_Network_UDP)
	If hServ = 0
		LogPrint ("Failed to open local UDP port (53)")
		InitOK = #False
	EndIf
	; create connection to real DNS server 
	Define hDNS = OpenNetworkConnection("8.8.8.8", 53, #PB_Network_UDP | #PB_Network_IPv4, 30)
	If hDNS = 0
		LogPrint ("Failed to connect DNS server")
		InitOK = #False
	EndIf
	
	; alloc memory for DNS messages
	Define *msg.DNS_HEADER = AllocateMemory(#UDP_MAX)
	Define bytes_read
	If *msg = 0
		LogPrint ("Failed to allocate network buffer (pf, 2kb only. WTF)")
		InitOK = #False
	EndIf
	
	; load filter DLL
	Define hDLLFilter = OpenLibrary(#PB_Any, "Filter.DLL")
	If IsLibrary(hDLLFilter)
		DLL_FilterQuery = GetFunction(hDLLFilter, "FilterQuery")
	EndIf
	If DLL_FilterQuery = 0
		LogPrint("Filter.DLL not loaded/unrecognized")
	EndIf
	
;}



; main
Repeat
	; window events
	Define E = WindowEvent()
	If E = #PB_Event_CloseWindow
		Break
	EndIf

	; if init failed, there is nothing to do
	If InitOK = #False
		If E = 0
			Delay(16)
		EndIf
		Continue
	EndIf

	; process network events
	Define Sleep = #False
	Define C = NetworkServerEvent(hServ)
	Define S = NetworkClientEvent(hDNS)
	
	; 1) Process DNS messages from clients (and route it to DNS server, or do something else)
	If C = #PB_NetworkEvent_Data
		; read received DNS message
		bytes_read = ReceiveNetworkData (EventClient(), *msg, #UDP_MAX)
		; process it
		DNSProcessQ (clients(), connections(), *msg, bytes_read, EventClient(), hDNS)
		; waste other received data for this event (if any and client is still alive)
		If FindMapElement(connections(), Str(EventClient())) And connections()\ref > 0
			While Not ReceiveNetworkData (EventClient(), *msg, #UDP_MAX) <= 0
			Wend
		EndIf
	Else
		Sleep = #True
	EndIf
	
	; 2) Process DNS messages from server (and send it back to corresponding client.. or do whatever)
	If S = #PB_NetworkEvent_Data
		; read received DNS message
		bytes_read = ReceiveNetworkData (hDNS, *msg, #UDP_MAX)
		; process it
		DNSProcessA (clients(), connections(), *msg, bytes_read)
		; waste other received data for this event (if any and client is still alive)
		If FindMapElement(connections(), Str(EventClient())) And connections()\ref > 0
			While Not ReceiveNetworkData (EventClient(), *msg, #UDP_MAX) <= 0
			Wend
		EndIf
	Else
		Sleep = #True
	EndIf
	
	
	; 3) Cls
	If Sleep
		; once per ~45 sec close client connections which are inactive for 30+ sec
		If Abs(Date() - Cycle) >= 45
			ForEach clients()
				; remove if client hangs for 30+ sec
				If Abs(Date() - clients()\date) >= 30
					; dec connection ref
					connections(Str(clients()\connection))\ref - 1
					If connections()\ref = 0
			 			CloseNetworkConnection(connections()\ID)
			 			DeleteMapElement(connections())
			 		EndIf
					DeleteMapElement(clients())
				EndIf
			Next
			Cycle = Date()
		Else
			; take some sleep if totally nothing happens
			If E = 0
				Delay(16)
			EndIf
		EndIf
	EndIf
ForEver



; cls
If IsLibrary(hDLLFilter)
	CloseLibrary(hDLLFilter)
EndIf
If hServ
	CloseNetworkServer(hServ)
EndIf
If hDNS
	CloseNetworkConnection(hDNS)
EndIf
If *msg
	FreeMemory(*msg)
EndIf

End
; IDE Options = PureBasic 5.60 (Windows - x86)
; CursorPosition = 406
; FirstLine = 223
; Folding = G908fv------
; EnableXP
; EnableOnError
; Executable = DNSProxy.exe
; CPU = 1
; SubSystem = /OPENGL
; CompileSourceDirectory
; EnablePurifier
; EnableBuildCount = 2
; EnableExeConstant
; IncludeVersionInfo
; VersionField0 = 1.0.0.%BUILDCOUNT
; VersionField1 = 1.0.0.%BUILDCOUNT
; VersionField2 = Luna Sole
; VersionField3 = DNS Proxy
; VersionField5 = 1.0.0.%BUILDCOUNT
; VersionField6 = 127.0.0.1 Firebox
; VersionField8 = %EXECUTABLE
; VersionField9 = Luna Sole
; VersionField14 = http://geocities.ws/lunasole